@@ -40,9 +40,11 @@
 #include "expand/esetinternal.h"
 #include "expand/expandweight.h"
 #include "exp10.h"
+#include "matcher/msetcmp.h"
 #include "matcher/multimatch.h"
 #include "omassert.h"
 #include "api/omenquireinternal.h"
+#include "roundestimate.h"
 #include "str.h"
 #include "weight/weightinternal.h"
 
@@ -78,6 +80,11 @@ RSet::operator=(const RSet &other)
     internal = other.internal;
 }
 
+RSet::RSet(RSet &&) = default;
+
+RSet&
+RSet::operator=(RSet &&) = default;
+
 RSet::~RSet()
 {
 }
@@ -175,6 +182,11 @@ MSet::operator=(const MSet &other)
     return *this;
 }
 
+MSet::MSet(MSet&&) = default;
+
+MSet&
+MSet::operator=(MSet&&) = default;
+
 void
 MSet::fetch_(Xapian::doccount first, Xapian::doccount last) const
 {
@@ -242,39 +254,9 @@ Xapian::doccount
 MSet::get_matches_estimated() const
 {
     Assert(internal.get() != 0);
-
-    // Doing this here avoids calculating if the estimate is never looked at,
-    // though does mean we recalculate if this method is called more than once.
-
-    Xapian::doccount m = internal->matches_lower_bound;
-    Xapian::doccount M = internal->matches_upper_bound;
-    Xapian::doccount e = internal->matches_estimated;
-
-    Xapian::doccount D = M - m;
-    if (D == 0 || e == 0) {
-	// Estimate is exact or zero.  A zero but non-exact estimate can happen
-	// with get_mset(0, 0).
-	return e;
-    }
-
-    Xapian::doccount r = Xapian::doccount(exp10(int(log10(D))) + 0.5);
-    while (r > e) r /= 10;
-
-    Xapian::doccount R = e / r * r;
-    if (R < m) {
-	R += r;
-    } else if (R > M) {
-	R -= r;
-    } else if (R < e && r % 2 == 0 && e - R == r / 2) {
-	// Round towards the centre of the range.
-	if (e - m < M - e) {
-	    R += r;
-	}
-    }
-
-    // If it all goes pear-shaped, just stick to the original estimate.
-    if (R < m || R > M) R = e;
-    return R;
+    return round_estimate(internal->matches_lower_bound,
+			  internal->matches_upper_bound,
+			  internal->matches_estimated);
 }
 
 Xapian::doccount
@@ -505,6 +487,16 @@ MSetIterator::get_collapse_count() const
     return mset.internal->items[index].collapse_count;
 }
 
+string
+MSetIterator::get_sort_key() const
+{
+    Assert(mset.internal.get());
+    Xapian::doccount size = mset.internal->items.size();
+    Xapian::doccount index = size - off_from_end;
+    AssertRel(index,<,size);
+    return mset.internal->items[index].sort_key;
+}
+
 string
 MSetIterator::get_description() const
 {
@@ -563,9 +555,9 @@ Enquire::Internal::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
     {
 	Xapian::doccount docs = db.get_doccount();
 	first = min(first, docs);
-	maxitems = min(maxitems, docs);
+	maxitems = min(maxitems, docs - first);
 	check_at_least = min(check_at_least, docs);
-	check_at_least = max(check_at_least, maxitems);
+	check_at_least = max(check_at_least, first + maxitems);
     }
 
     AutoPtr<Xapian::Weight::Internal> stats(new Xapian::Weight::Internal);
@@ -776,6 +768,11 @@ Enquire::operator=(const Enquire & other)
     internal = other.internal;
 }
 
+Enquire::Enquire(Enquire&&) = default;
+
+Enquire&
+Enquire::operator=(Enquire&&) = default;
+
 Enquire::Enquire(const Database &databases)
     : internal(new Internal(databases))
 {
