@@ -1,7 +1,7 @@
-/** @file remote-database.cc
+/** @file
  *  @brief Remote backend database class
  */
-/* Copyright (C) 2006,2007,2008,2009,2010,2015 Olly Betts
+/* Copyright (C) 2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2019,2020 Olly Betts
  * Copyright (C) 2007,2009,2010 Lemur Consulting Ltd
  *
  * This program is free software; you can redistribute it and/or
@@ -23,31 +23,57 @@
 
 #include "remote-database.h"
 
-#include "safeerrno.h"
 #include <signal.h>
 
 #include "autoptr.h"
-#include "emptypostlist.h"
-#include "inmemory_positionlist.h"
-#include "net/length.h"
+#include "backends/inmemory/inmemory_positionlist.h"
 #include "net_postlist.h"
 #include "net_termlist.h"
+#include "noreturn.h"
 #include "remote-document.h"
 #include "omassert.h"
 #include "realtime.h"
-#include "serialise.h"
+#include "net/length.h"
+#include "net/serialise.h"
+#include "net/serialise-error.h"
 #include "serialise-double.h"
 #include "str.h"
 #include "stringutils.h" // For STRINGIZE().
-#include "weightinternal.h"
+#include "weight/weightinternal.h"
 
+#include <cerrno>
 #include <string>
 #include <vector>
 
+#include "xapian/constants.h"
 #include "xapian/error.h"
 #include "xapian/matchspy.h"
 
 using namespace std;
+using Xapian::Internal::intrusive_ptr;
+
+/// Return true if further replies should be expected.
+static inline bool
+is_intermediate_reply(int msg_code, int reply_code)
+{
+    return reply_code == REPLY_ALLTERMS ||
+	   reply_code == REPLY_DOCDATA ||
+	   reply_code == REPLY_VALUE ||
+	   reply_code == REPLY_TERMLIST ||
+	   reply_code == REPLY_POSITIONLIST ||
+	   reply_code == REPLY_POSTLISTSTART ||
+	   reply_code == REPLY_POSTLISTITEM ||
+	   reply_code == REPLY_METADATAKEYLIST ||
+	   (msg_code == MSG_TERMLIST && reply_code == REPLY_DOCLENGTH);
+}
+
+XAPIAN_NORETURN(static void throw_handshake_failed(const string & context));
+static void
+throw_handshake_failed(const string & context)
+{
+    throw Xapian::NetworkError("Handshake failed - is this a Xapian server?",
+			       context);
+}
 
 XAPIAN_NORETURN(static void throw_connection_closed_unexpectedly());
 static void
@@ -57,7 +83,8 @@ throw_connection_closed_unexpectedly()
 }
 
 RemoteDatabase::RemoteDatabase(int fd, double timeout_,
-			       const string & context_, bool writable)
+			       const string & context_, bool writable,
+			       int flags)
 	: link(fd, fd, context_),
 	  context(context_),
 	  cached_stats_valid(),
@@ -82,54 +109,16 @@ RemoteDatabase::RemoteDatabase(int fd, double timeout_,
 	transaction_state = TRANSACTION_UNIMPLEMENTED;
     }
 
-    string message;
-    char type = get_message(message);
+    update_stats(MSG_MAX);
 
-    if (reply_type(type) != REPLY_GREETING || message.size() < 3) {
-	if (type == 'O' && message.size() == size_t('M') && message[0] == ' ') {
-	    // The server reply used to start "OM ", which will now be
-	    // interpreted as a type 'O' message of length size_t('M')
-	    // with first character ' '.
-	    throw Xapian::NetworkError("Server protocol version too old", context);
+    if (writable) {
+	if (flags & Xapian::DB_RETRY_LOCK) {
+	    const string & body = encode_length(flags & Xapian::DB_RETRY_LOCK);
+	    update_stats(MSG_WRITEACCESS, body);
+	} else {
+	    update_stats(MSG_WRITEACCESS);
 	}
-	throw Xapian::NetworkError("Handshake failed - is this a Xapian server?", context);
     }
-
-    const char *p = message.c_str();
-    const char *p_end = p + message.size();
-
-    // The protocol major versions must match.  The protocol minor version of
-    // the server must be >= that of the client.
-    int protocol_major = static_cast<unsigned char>(*p++);
-    int protocol_minor = static_cast<unsigned char>(*p++);
-    if (protocol_major != XAPIAN_REMOTE_PROTOCOL_MAJOR_VERSION ||
-	protocol_minor < XAPIAN_REMOTE_PROTOCOL_MINOR_VERSION) {
-	string errmsg("Server supports protocol version");
-	if (protocol_minor) {
-	    errmsg += "s ";
-	    errmsg += str(protocol_major);
-	    errmsg += ".0 to ";
-	}
-	errmsg += str(protocol_major);
-	errmsg += '.';
-	errmsg += str(protocol_minor);
-	errmsg +=
-	    " - client is using "
-	    STRINGIZE(XAPIAN_REMOTE_PROTOCOL_MAJOR_VERSION)
-	    "."
-	    STRINGIZE(XAPIAN_REMOTE_PROTOCOL_MINOR_VERSION);
-	throw Xapian::NetworkError(errmsg, context);
-    }
-
-    apply_stats_update(p, p_end);
-
-    if (writable) update_stats(MSG_WRITEACCESS);
-}
-
-RemoteDatabase *
-RemoteDatabase::as_remotedatabase()
-{
-    return this;
 }
 
 void
@@ -151,19 +140,18 @@ RemoteDatabase::open_metadata_keylist(const std::string &prefix) const
     string message;
     AutoPtr<NetworkTermList> tlist(
 	new NetworkTermList(0, doccount,
-			    Xapian::Internal::RefCntPtr<const RemoteDatabase>(this),
+			    intrusive_ptr<const RemoteDatabase>(this),
 			    0));
     vector<NetworkTermListItem> & items = tlist->items;
 
-    char type;
-    while ((type = get_message(message)) == REPLY_METADATAKEYLIST) {
+    string term = prefix;
+    while (get_message_or_done(message, REPLY_METADATAKEYLIST)) {
 	NetworkTermListItem item;
-	item.tname = message;
+	term.resize(size_t(static_cast<unsigned char>(message[0])));
+	term.append(message, 1, string::npos);
+	item.tname = term;
 	items.push_back(item);
     }
-    if (type != REPLY_DONE) {
-	throw Xapian::NetworkError("Bad message received", context);
-    }
 
     tlist->current_position = tlist->items.begin();
     return tlist.release();
@@ -191,23 +179,22 @@ RemoteDatabase::open_term_list(Xapian::docid did) const
 
     AutoPtr<NetworkTermList> tlist(
 	new NetworkTermList(doclen, doccount,
-			    Xapian::Internal::RefCntPtr<const RemoteDatabase>(this),
+			    intrusive_ptr<const RemoteDatabase>(this),
 			    did));
     vector<NetworkTermListItem> & items = tlist->items;
 
-    char type;
-    while ((type = get_message(message)) == REPLY_TERMLIST) {
+    string term;
+    while (get_message_or_done(message, REPLY_TERMLIST)) {
 	NetworkTermListItem item;
 	p = message.data();
 	p_end = p + message.size();
 	decode_length(&p, p_end, item.wdf);
 	decode_length(&p, p_end, item.termfreq);
-	item.tname.assign(p, p_end);
+	term.resize(size_t(static_cast<unsigned char>(*p++)));
+	term.append(p, p_end);
+	item.tname = term;
 	items.push_back(item);
     }
-    if (type != REPLY_DONE) {
-	throw Xapian::NetworkError("Bad message received", context);
-    }
 
     tlist->current_position = tlist->items.begin();
     return tlist.release();
@@ -222,23 +209,22 @@ RemoteDatabase::open_allterms(const string & prefix) const {
 
     AutoPtr<NetworkTermList> tlist(
 	new NetworkTermList(0, doccount,
-			    Xapian::Internal::RefCntPtr<const RemoteDatabase>(this),
+			    intrusive_ptr<const RemoteDatabase>(this),
 			    0));
     vector<NetworkTermListItem> & items = tlist->items;
 
+    string term = prefix;
     string message;
-    char type;
-    while ((type = get_message(message)) == REPLY_ALLTERMS) {
+    while (get_message_or_done(message, REPLY_ALLTERMS)) {
 	NetworkTermListItem item;
 	const char * p = message.data();
 	const char * p_end = p + message.size();
 	decode_length(&p, p_end, item.termfreq);
-	item.tname.assign(p, p_end);
+	term.resize(size_t(static_cast<unsigned char>(*p++)));
+	term.append(p, p_end);
+	item.tname = term;
 	items.push_back(item);
     }
-    if (type != REPLY_DONE) {
-	throw Xapian::NetworkError("Bad message received", context);
-    }
 
     tlist->current_position = tlist->items.begin();
     return tlist.release();
@@ -247,7 +233,7 @@ RemoteDatabase::open_allterms(const string & prefix) const {
 LeafPostList *
 RemoteDatabase::open_post_list(const string &term) const
 {
-    return new NetworkPostList(Xapian::Internal::RefCntPtr<const RemoteDatabase>(this), term);
+    return new NetworkPostList(intrusive_ptr<const RemoteDatabase>(this), term);
 }
 
 Xapian::doccount
@@ -256,7 +242,6 @@ RemoteDatabase::read_post_list(const string &term, NetworkPostList & pl) const
     send_message(MSG_POSTLIST, term);
 
     string message;
-    char type;
     get_message(message, REPLY_POSTLISTSTART);
 
     const char * p = message.data();
@@ -264,12 +249,9 @@ RemoteDatabase::read_post_list(const string &term, NetworkPostList & pl) const
     Xapian::doccount termfreq;
     decode_length(&p, p_end, termfreq);
 
-    while ((type = get_message(message)) == REPLY_POSTLISTITEM) {
+    while (get_message_or_done(message, REPLY_POSTLISTITEM)) {
 	pl.append_posting(message);
     }
-    if (type != REPLY_DONE) {
-	throw Xapian::NetworkError("Bad message received", context);
-    }
 
     return termfreq;
 }
@@ -282,9 +264,8 @@ RemoteDatabase::open_position_list(Xapian::docid did, const string &term) const
     vector<Xapian::termpos> positions;
 
     string message;
-    char type;
     Xapian::termpos lastpos = static_cast<Xapian::termpos>(-1);
-    while ((type = get_message(message)) == REPLY_POSITIONLIST) {
+    while (get_message_or_done(message, REPLY_POSITIONLIST)) {
 	const char * p = message.data();
 	const char * p_end = p + message.size();
 	Xapian::termpos inc;
@@ -292,9 +273,6 @@ RemoteDatabase::open_position_list(Xapian::docid did, const string &term) const
 	lastpos += inc + 1;
 	positions.push_back(lastpos);
     }
-    if (type != REPLY_DONE) {
-	throw Xapian::NetworkError("Bad message received", context);
-    }
 
     return new InMemoryPositionList(positions);
 }
@@ -306,11 +284,11 @@ RemoteDatabase::has_positions() const
     return has_positional_info;
 }
 
-void
+bool
 RemoteDatabase::reopen()
 {
-    update_stats(MSG_REOPEN);
     mru_slot = Xapian::BAD_VALUENO;
+    return update_stats(MSG_REOPEN);
 }
 
 void
@@ -319,13 +297,28 @@ RemoteDatabase::close()
     do_close();
 }
 
-// Currently lazy is used when fetching documents from the MSet, and in three
-// cases in multimatch.cc.  One of the latter is when using a MatchDecider,
-// which we don't support with the remote backend currently.  The others are
-// for the sort key and collapse key which in the remote cases are fetched
-// during the remote match and passed across with the MSet.  So we can safely
-// ignore "lazy" here for now without any performance penalty during the match
-// process.
+// Currently lazy is used:
+//
+// * To implement API flag Xapian::DOC_ASSUME_VALID which can be specified when
+//   calling method Database::get_document()
+//
+// * To read values for backends without streamed values in SlowValueList
+//
+// * If you call get_data(), values_begin() or values_count() on a Document
+//   object passed to a KeyMaker, MatchDecider, MatchSpy during the match
+//
+// The first is relevant to the remote backend, but doesn't happen during
+// the match.
+//
+// SlowValueList is used with the remote backend, but not to read values
+// during the match.
+//
+// KeyMaker and MatchSpy happens on the server with the remote backend, so
+// they aren't relevant here.
+//
+// So the cases which are relevant to the remote backend don't matter during
+// the match, and so we can ignore the lazy flag here without affecting matcher
+// performance.
 Xapian::Document::Internal *
 RemoteDatabase::open_document(Xapian::docid did, bool /*lazy*/) const
 {
@@ -336,40 +329,69 @@ RemoteDatabase::open_document(Xapian::docid did, bool /*lazy*/) const
     map<Xapian::valueno, string> values;
     get_message(doc_data, REPLY_DOCDATA);
 
-    reply_type type;
     string message;
-    while ((type = get_message(message)) == REPLY_VALUE) {
+    while (get_message_or_done(message, REPLY_VALUE)) {
 	const char * p = message.data();
 	const char * p_end = p + message.size();
 	Xapian::valueno slot;
 	decode_length(&p, p_end, slot);
 	values.insert(make_pair(slot, string(p, p_end)));
     }
-    if (type != REPLY_DONE) {
-	throw Xapian::NetworkError("Bad message received", context);
-    }
 
     return new RemoteDocument(this, did, doc_data, values);
 }
 
-void
-RemoteDatabase::update_stats(message_type msg_code) const
+bool
+RemoteDatabase::update_stats(message_type msg_code, const string & body) const
 {
-    send_message(msg_code, string());
+    // MSG_MAX signals that we're handling the opening greeting, which isn't in
+    // response to an explicit message.
+    if (msg_code != MSG_MAX)
+	send_message(msg_code, body);
+
     string message;
-    get_message(message, REPLY_UPDATE);
-    const char * p = message.c_str();
-    const char * p_end = p + message.size();
-    apply_stats_update(p, p_end);
-}
+    if (!get_message_or_done(message, REPLY_UPDATE)) {
+	// The database was already open at the latest revision.
+	return false;
+    }
+
+    if (message.size() < 3) {
+	throw_handshake_failed(context);
+    }
+    const char *p = message.c_str();
+    const char *p_end = p + message.size();
+
+    // The protocol major versions must match.  The protocol minor version of
+    // the server must be >= that of the client.
+    int protocol_major = static_cast<unsigned char>(*p++);
+    int protocol_minor = static_cast<unsigned char>(*p++);
+    if (protocol_major != XAPIAN_REMOTE_PROTOCOL_MAJOR_VERSION ||
+	protocol_minor < XAPIAN_REMOTE_PROTOCOL_MINOR_VERSION) {
+	string errmsg("Server supports protocol version");
+	if (protocol_minor) {
+	    errmsg += "s ";
+	    errmsg += str(protocol_major);
+	    errmsg += ".0 to ";
+	} else {
+	    errmsg += ' ';
+	}
+	errmsg += str(protocol_major);
+	errmsg += '.';
+	errmsg += str(protocol_minor);
+	errmsg +=
+	    " - client is using "
+	    STRINGIZE(XAPIAN_REMOTE_PROTOCOL_MAJOR_VERSION)
+	    "."
+	    STRINGIZE(XAPIAN_REMOTE_PROTOCOL_MINOR_VERSION);
+	throw Xapian::NetworkError(errmsg, context);
+    }
 
-void
-RemoteDatabase::apply_stats_update(const char * p, const char * p_end) const
-{
     decode_length(&p, p_end, doccount);
     decode_length(&p, p_end, lastdocid);
+    lastdocid += doccount;
     decode_length(&p, p_end, doclen_lbound);
     decode_length(&p, p_end, doclen_ubound);
+    doclen_ubound += doclen_lbound;
     if (p == p_end) {
 	throw Xapian::NetworkError("Bad stats update message received", context);
     }
@@ -377,6 +399,7 @@ RemoteDatabase::apply_stats_update(const char * p, const char * p_end) const
     decode_length(&p, p_end, total_length);
     uuid.assign(p, p_end);
     cached_stats_valid = true;
+    return true;
 }
 
 Xapian::doccount
@@ -393,63 +416,56 @@ RemoteDatabase::get_lastdocid() const
     return lastdocid;
 }
 
-totlen_t
+Xapian::totallength
 RemoteDatabase::get_total_length() const
 {
     if (!cached_stats_valid) update_stats();
     return total_length;
 }
 
-Xapian::doclength
-RemoteDatabase::get_avlength() const
-{
-    if (!cached_stats_valid) update_stats();
-    if (rare(doccount == 0)) return 0;
-    return Xapian::doclength(total_length) / doccount;
-}
-
 bool
 RemoteDatabase::term_exists(const string & tname) const
 {
     Assert(!tname.empty());
     send_message(MSG_TERMEXISTS, tname);
     string message;
-    reply_type type = get_message(message);
-    if (type != REPLY_TERMEXISTS && type != REPLY_TERMDOESNTEXIST) {
-	throw Xapian::NetworkError("Bad message received", context);
-    }
+    reply_type type = get_message(message,
+				  REPLY_TERMEXISTS,
+				  REPLY_TERMDOESNTEXIST);
     return (type == REPLY_TERMEXISTS);
 }
 
-Xapian::doccount
-RemoteDatabase::get_termfreq(const string & tname) const
+void
+RemoteDatabase::get_freqs(const string & term,
+			  Xapian::doccount * termfreq_ptr,
+			  Xapian::termcount * collfreq_ptr) const
 {
-    Assert(!tname.empty());
-    send_message(MSG_TERMFREQ, tname);
+    Assert(!term.empty());
     string message;
-    get_message(message, REPLY_TERMFREQ);
-    const char * p = message.data();
-    const char * p_end = p + message.size();
-    Xapian::doccount r;
-    decode_length(&p, p_end, r);
-    return r;
+    const char * p;
+    const char * p_end;
+    if (termfreq_ptr) {
+	if (collfreq_ptr) {
+	    send_message(MSG_FREQS, term);
+	    get_message(message, REPLY_FREQS);
+	} else {
+	    send_message(MSG_TERMFREQ, term);
+	    get_message(message, REPLY_TERMFREQ);
+	}
+	p = message.data();
+	p_end = p + message.size();
+	decode_length(&p, p_end, *termfreq_ptr);
+    } else if (collfreq_ptr) {
+	send_message(MSG_COLLFREQ, term);
+	get_message(message, REPLY_COLLFREQ);
+	p = message.data();
+	p_end = p + message.size();
+    }
+    if (collfreq_ptr) {
+	decode_length(&p, p_end, *collfreq_ptr);
+    }
 }
 
-Xapian::termcount
-RemoteDatabase::get_collection_freq(const string & tname) const
-{
-    Assert(!tname.empty());
-    send_message(MSG_COLLFREQ, tname);
-    string message;
-    get_message(message, REPLY_COLLFREQ);
-    const char * p = message.data();
-    const char * p_end = p + message.size();
-    Xapian::termcount r;
-    decode_length(&p, p_end, r);
-    return r;
-}
-
-
 void
 RemoteDatabase::read_value_stats(Xapian::valueno slot) const
 {
@@ -534,33 +550,81 @@ RemoteDatabase::get_doclength(Xapian::docid did) const
     return doclen;
 }
 
+Xapian::termcount
+RemoteDatabase::get_unique_terms(Xapian::docid did) const
+{
+    Assert(did != 0);
+    send_message(MSG_UNIQUETERMS, encode_length(did));
+    string message;
+    get_message(message, REPLY_UNIQUETERMS);
+    const char * p = message.c_str();
+    const char * p_end = p + message.size();
+    Xapian::termcount doclen;
+    decode_length(&p, p_end, doclen);
+    if (p != p_end) {
+	throw Xapian::NetworkError("Bad REPLY_UNIQUETERMS message received", context);
+    }
+    return doclen;
+}
+
 reply_type
-RemoteDatabase::get_message(string &result, reply_type required_type) const
+RemoteDatabase::get_message(string &result,
+			    reply_type required_type,
+			    reply_type required_type2) const
 {
     double end_time = RealTime::end_time(timeout);
-    int type_int = link.get_message(result, end_time);
-    if (type_int < 0)
+    int type = link.get_message(result, end_time);
+    if (pending_reply >= 0 && !is_intermediate_reply(pending_reply, type)) {
+	pending_reply = -1;
+    }
+    if (type < 0)
 	throw_connection_closed_unexpectedly();
-    reply_type type = static_cast<reply_type>(type_int);
+    if (rare(type) >= REPLY_MAX) {
+	if (required_type == REPLY_UPDATE)
+	    throw_handshake_failed(context);
+	string errmsg("Invalid reply type ");
+	errmsg += str(type);
+	throw Xapian::NetworkError(errmsg);
+    }
     if (type == REPLY_EXCEPTION) {
 	unserialise_error(result, "REMOTE:", context);
     }
-    if (required_type != REPLY_MAX && type != required_type) {
+    if (type != required_type && type != required_type2) {
 	string errmsg("Expecting reply type ");
 	errmsg += str(int(required_type));
+	if (required_type2 != required_type) {
+	    errmsg += " or ";
+	    errmsg += str(int(required_type2));
+	}
 	errmsg += ", got ";
-	errmsg += str(int(type));
+	errmsg += str(type);
 	throw Xapian::NetworkError(errmsg);
     }
 
-    return type;
+    return static_cast<reply_type>(type);
 }
 
 void
 RemoteDatabase::send_message(message_type type, const string &message) const
 {
     double end_time = RealTime::end_time(timeout);
+    while (pending_reply >= 0) {
+	string dummy;
+	int reply_code = link.get_message(dummy, end_time);
+	if (reply_code < 0)
+	    throw_connection_closed_unexpectedly();
+	if (!is_intermediate_reply(pending_reply, reply_code)) {
+	    pending_reply = -1;
+	}
+    }
     link.send_message(static_cast<unsigned char>(type), message, end_time);
+    if (type == MSG_REMOVESPELLING) {
+	// MSG_REMOVESPELLING is the only message we send which doesn't expect
+	// a reply (except MSG_SHUTDOWN which causes us to exit anyway).
+	pending_reply = -1;
+    } else {
+	pending_reply = int(type);
+    }
 }
 
 void
@@ -571,31 +635,47 @@ RemoteDatabase::do_close()
     // it here.
     bool writable = (transaction_state != TRANSACTION_UNIMPLEMENTED);
 
-    // Only call dtor_called() if we're writable.
-    if (writable) dtor_called();
+    if (writable) {
+	try {
+	    if (transaction_active()) {
+		cancel_transaction();
+	    } else {
+		commit();
+	    }
+	} catch (...) {
+	    try {
+		link.do_close();
+	    } catch (...) {
+	    }
+	    throw;
+	}
 
-    // If we're writable, wait for a confirmation of the close, so we know that
-    // changes have been written and flushed, and the database write lock
-    // released.  For the non-writable case, there's no need to wait, so don't
-    // slow down searching by waiting here.
-    link.do_close(writable);
+	// If we're writable, send a shutdown message to the server and wait
+	// for it to close its end of the connection so we know that changes
+	// have been written and flushed, and the database write lock released.
+	// For the non-writable case, there's no need to wait - it would just
+	// slow down searching needlessly.
+	link.shutdown();
+    }
+    link.do_close();
 }
 
 void
-RemoteDatabase::set_query(const Xapian::Query::Internal *query,
-			 Xapian::termcount qlen,
-			 Xapian::doccount collapse_max,
-			 Xapian::valueno collapse_key,
-			 Xapian::Enquire::docid_order order,
-			 Xapian::valueno sort_key,
-			 Xapian::Enquire::Internal::sort_setting sort_by,
-			 bool sort_value_forward,
-			 int percent_cutoff, Xapian::weight weight_cutoff,
-			 const Xapian::Weight *wtscheme,
-			 const Xapian::RSet &omrset,
-			 const vector<Xapian::MatchSpy *> & matchspies)
+RemoteDatabase::set_query(const Xapian::Query& query,
+			  Xapian::termcount qlen,
+			  Xapian::doccount collapse_max,
+			  Xapian::valueno collapse_key,
+			  Xapian::Enquire::docid_order order,
+			  Xapian::valueno sort_key,
+			  Xapian::Enquire::Internal::sort_setting sort_by,
+			  bool sort_value_forward,
+			  double time_limit,
+			  int percent_cutoff, double weight_cutoff,
+			  const Xapian::Weight *wtscheme,
+			  const Xapian::RSet &omrset,
+			  const vector<Xapian::Internal::opt_intrusive_ptr<Xapian::MatchSpy>> & matchspies)
 {
-    string tmp = query->serialise();
+    string tmp = query.serialise();
     string message = encode_length(tmp.size());
     message += tmp;
 
@@ -607,6 +687,7 @@ RemoteDatabase::set_query(const Xapian::Query::Internal *query,
     message += encode_length(sort_key);
     message += char('0' + sort_by);
     message += char('0' + sort_value_forward);
+    message += serialise_double(time_limit);
     message += char(percent_cutoff);
     message += serialise_double(weight_cutoff);
 
@@ -622,21 +703,20 @@ RemoteDatabase::set_query(const Xapian::Query::Internal *query,
     message += encode_length(tmp.size());
     message += tmp;
 
-    vector<Xapian::MatchSpy *>::const_iterator i;
-    for (i = matchspies.begin(); i != matchspies.end(); ++i) {
-	tmp = (*i)->name();
+    for (auto i : matchspies) {
+	tmp = i->name();
 	if (tmp.empty()) {
 	    throw Xapian::UnimplementedError("MatchSpy subclass not suitable for use with remote searches - name() method returned empty string");
 	}
 	message += encode_length(tmp.size());
 	message += tmp;
 
-	tmp = (*i)->serialise();
+	tmp = i->serialise();
 	message += encode_length(tmp.size());
 	message += tmp;
     }
 
-    send_message(MSG_QUERY_NEW, message);
+    send_message(MSG_QUERY, message);
 }
 
 bool
@@ -646,7 +726,8 @@ RemoteDatabase::get_remote_stats(bool nowait, Xapian::Weight::Internal &out)
 
     string message;
     get_message(message, REPLY_STATS);
-    out = unserialise_stats(message);
+    const char* p = message.data();
+    unserialise_stats(p, p + message.size(), out);
 
     return true;
 }
@@ -666,43 +747,52 @@ RemoteDatabase::send_global_stats(Xapian::doccount first,
 
 void
 RemoteDatabase::get_mset(Xapian::MSet &mset,
-			 const vector<Xapian::MatchSpy *> & matchspies)
+			 const vector<Xapian::Internal::opt_intrusive_ptr<Xapian::MatchSpy>> & matchspies)
 {
     string message;
-    get_message(message, REPLY_RESULTS_NEW);
+    get_message(message, REPLY_RESULTS);
     const char * p = message.data();
     const char * p_end = p + message.size();
 
-    vector<Xapian::MatchSpy *>::const_iterator i;
-    for (i = matchspies.begin(); i != matchspies.end(); ++i) {
+    for (auto i : matchspies) {
 	if (p == p_end)
 	    throw Xapian::NetworkError("Expected serialised matchspy");
 	size_t len;
 	decode_length_and_check(&p, p_end, len);
 	string spyresults(p, len);
 	p += len;
-	(*i)->merge_results(spyresults);
+	i->merge_results(spyresults);
     }
-    mset = unserialise_mset_new(p, p_end);
+    mset = unserialise_mset(p, p_end);
 }
 
 void
 RemoteDatabase::commit()
 {
+    if (!uncommitted_changes) return;
+
     send_message(MSG_COMMIT, string());
 
     // We need to wait for a response to ensure documents have been committed.
     string message;
     get_message(message, REPLY_DONE);
+
+    uncommitted_changes = false;
 }
 
 void
 RemoteDatabase::cancel()
 {
+    if (!uncommitted_changes) return;
+
     cached_stats_valid = false;
     mru_slot = Xapian::BAD_VALUENO;
 
     send_message(MSG_CANCEL, string());
+    string dummy;
+    get_message(dummy, REPLY_DONE);
+
+    uncommitted_changes = false;
 }
 
 Xapian::docid
@@ -710,6 +800,7 @@ RemoteDatabase::add_document(const Xapian::Document & doc)
 {
     cached_stats_valid = false;
     mru_slot = Xapian::BAD_VALUENO;
+    uncommitted_changes = true;
 
     send_message(MSG_ADDDOCUMENT, serialise_document(doc));
 
@@ -728,6 +819,7 @@ RemoteDatabase::delete_document(Xapian::docid did)
 {
     cached_stats_valid = false;
     mru_slot = Xapian::BAD_VALUENO;
+    uncommitted_changes = true;
 
     send_message(MSG_DELETEDOCUMENT, encode_length(did));
     string dummy;
@@ -739,8 +831,11 @@ RemoteDatabase::delete_document(const std::string & unique_term)
 {
     cached_stats_valid = false;
     mru_slot = Xapian::BAD_VALUENO;
+    uncommitted_changes = true;
 
     send_message(MSG_DELETEDOCUMENTTERM, unique_term);
+    string dummy;
+    get_message(dummy, REPLY_DONE);
 }
 
 void
@@ -749,11 +844,14 @@ RemoteDatabase::replace_document(Xapian::docid did,
 {
     cached_stats_valid = false;
     mru_slot = Xapian::BAD_VALUENO;
+    uncommitted_changes = true;
 
     string message = encode_length(did);
     message += serialise_document(doc);
 
     send_message(MSG_REPLACEDOCUMENT, message);
+    string dummy;
+    get_message(dummy, REPLY_DONE);
 }
 
 Xapian::docid
@@ -762,6 +860,7 @@ RemoteDatabase::replace_document(const std::string & unique_term,
 {
     cached_stats_valid = false;
     mru_slot = Xapian::BAD_VALUENO;
+    uncommitted_changes = true;
 
     string message = encode_length(unique_term.size());
     message += unique_term;
@@ -796,26 +895,42 @@ RemoteDatabase::get_metadata(const string & key) const
 void
 RemoteDatabase::set_metadata(const string & key, const string & value)
 {
+    uncommitted_changes = true;
+
     string data = encode_length(key.size());
     data += key;
     data += value;
     send_message(MSG_SETMETADATA, data);
+    string dummy;
+    get_message(dummy, REPLY_DONE);
 }
 
 void
 RemoteDatabase::add_spelling(const string & word,
 			     Xapian::termcount freqinc) const
 {
+    uncommitted_changes = true;
+
     string data = encode_length(freqinc);
     data += word;
     send_message(MSG_ADDSPELLING, data);
+    string dummy;
+    get_message(dummy, REPLY_DONE);
 }
 
 void
 RemoteDatabase::remove_spelling(const string & word,
 				Xapian::termcount freqdec) const
 {
+    uncommitted_changes = true;
+
     string data = encode_length(freqdec);
     data += word;
     send_message(MSG_REMOVESPELLING, data);
 }
+
+bool
+RemoteDatabase::locked() const
+{
+    throw Xapian::UnimplementedError("Database::locked() not implemented for remote backend");
+}
